import enum
from datetime import datetime
from typing import Annotated, Any, Literal, NewType, Union

import pydantic

# region API Models


class Response[T](pydantic.BaseModel):
    success: bool
    data: T


class ListResponse[T](pydantic.BaseModel):
    object: Literal["list"]
    data: list[T]


class DeleteResponse(pydantic.BaseModel, extra="forbid"):
    success: bool


# endregion

# region Lock / Unlock Models


class ActionResponse(pydantic.BaseModel, extra="forbid"):
    noColor: bool
    object: str
    title: str
    message: str | None


class LockResponse(ActionResponse, extra="forbid"):
    pass


class UnlockResponse(ActionResponse, extra="forbid"):
    raw: str


class SyncResponse(ActionResponse, extra="forbid"):
    pass


class UnlockPayload(pydantic.BaseModel):
    password: pydantic.SecretStr

    @pydantic.field_serializer("password", when_used="json")
    def serialize_password(self, password: pydantic.SecretStr) -> str:
        return password.get_secret_value()


class GeneratePasswordResponse(pydantic.BaseModel):
    object: Literal["string"]
    data: pydantic.SecretStr


# endregion

# region Folder Models

FolderID = NewType("FolderID", str)


class Folder(pydantic.BaseModel):
    object: Literal["folder"] = pydantic.Field(exclude=True)
    name: str
    id: FolderID | None = pydantic.Field(exclude=True)


class FolderNew(pydantic.BaseModel):
    name: str


# endregion

# region Collection Models

CollectionID = NewType("CollectionID", str)

# endregion

# region Item Models

ItemID = NewType("ItemID", str)
OrgID = NewType("OrgID", str)


class ItemType(enum.IntEnum):
    login = 1
    secure_note = 2
    card = 3
    identity = 4


class URIMatch(enum.IntEnum):
    base_domain = 0
    host = 1
    starts_with = 2
    exact = 3
    regex = 4
    never = 5


class FieldType(enum.IntEnum):
    text = 0
    hidden = 1
    checkbox = 2
    linked = 3


class LinkedType(enum.IntEnum):
    username = 100
    password = 101


class UriMatch(pydantic.BaseModel, extra="forbid"):
    match: URIMatch | None
    uri: str


class PasswordHistory(pydantic.BaseModel, extra="forbid"):
    last_used: datetime = pydantic.Field(alias="lastUsedDate")
    password: pydantic.SecretStr

    @pydantic.field_serializer("password", when_used="json")
    def serialize_secretstr(self, value: pydantic.SecretStr | None) -> str | None:
        if value is None:
            return None
        return value.get_secret_value()


class FieldText(pydantic.BaseModel, extra="forbid"):
    type: Literal[FieldType.text] = pydantic.Field(default=FieldType.text)
    name: str
    value: str
    linkedId: None = None


class FieldHidden(pydantic.BaseModel, extra="forbid"):
    type: Literal[FieldType.hidden] = pydantic.Field(default=FieldType.hidden)
    name: str
    value: pydantic.SecretStr
    linkedId: None = None

    @pydantic.field_serializer("value", when_used="json")
    def serialize_secretstr(self, value: pydantic.SecretStr | None) -> str | None:
        if value is None:
            return None
        return value.get_secret_value()


class FieldCheckbox(pydantic.BaseModel, extra="forbid"):
    type: Literal[FieldType.checkbox] = pydantic.Field(default=FieldType.checkbox)
    name: str
    value: bool
    linkedId: None = None


class FieldLinked(pydantic.BaseModel, extra="forbid"):
    type: Literal[FieldType.linked] = pydantic.Field(default=FieldType.linked)
    name: str
    value: None = None
    linkedId: LinkedType


Fields = Annotated[Union[FieldText, FieldHidden, FieldCheckbox, FieldLinked], pydantic.Field(discriminator="type")]


class ItemLoginData(pydantic.BaseModel, extra="forbid"):
    uris: list[UriMatch] | None = None
    username: str | None = None
    password: pydantic.SecretStr | None = None
    totp: str | None = None
    passwordRevisionDate: datetime | None = pydantic.Field(default=None, alias="passwordRevisionDate", exclude=True)

    @pydantic.field_serializer("password", when_used="json")
    def serialize_secretstr(self, value: pydantic.SecretStr | None) -> str | None:
        if value is None:
            return None
        return value.get_secret_value()


class ItemLogin(pydantic.BaseModel, extra="forbid"):
    object: Literal["item"] = pydantic.Field(exclude=True)
    type: Literal[ItemType.login]
    id: ItemID = pydantic.Field(exclude=True)
    folder_id: FolderID | None = pydantic.Field(alias="folderId")
    organization_id: OrgID | None = pydantic.Field(alias="organizationId")
    collection_ids: list[CollectionID] | None = pydantic.Field(default=None, alias="collectionIds")
    attachments: list[Any] = pydantic.Field(alias="attachments", default_factory=list[Any])
    creation_date: datetime = pydantic.Field(alias="creationDate", exclude=True)
    revision_date: datetime = pydantic.Field(alias="revisionDate", exclude=True)
    deleted_date: datetime | None = pydantic.Field(alias="deletedDate", exclude=True)
    name: str
    login: ItemLoginData
    notes: str | None
    fields: list[Fields] | None = None
    reprompt: bool
    favorite: bool
    password_history: list[PasswordHistory] | None = pydantic.Field(alias="passwordHistory", exclude=True)

    @pydantic.field_serializer("reprompt", when_used="json")
    def serialize_reprompt(self, value: bool) -> int:
        return 1 if value else 0

    @pydantic.field_validator("reprompt", mode="before")
    def validate_reprompt(cls, value: int | bool) -> bool:
        if isinstance(value, bool):
            return value
        return value == 1


class ItemSecureNote(pydantic.BaseModel, extra="forbid"):
    object: Literal["item"] = pydantic.Field(exclude=True)
    type: Literal[ItemType.secure_note] = pydantic.Field(exclude=True)
    secureNote: dict[str, Any] = pydantic.Field(alias="secureNote")
    id: ItemID = pydantic.Field(exclude=True)
    folder_id: FolderID | None = pydantic.Field(alias="folderId")
    organization_id: OrgID | None = pydantic.Field(alias="organizationId")
    attachments: list[Any] = pydantic.Field(alias="attachments", default_factory=list[Any])
    collection_ids: list[CollectionID] | None = pydantic.Field(alias="collectionIds")
    creation_date: datetime = pydantic.Field(alias="creationDate")
    revision_date: datetime = pydantic.Field(alias="revisionDate")
    deleted_date: datetime | None = pydantic.Field(alias="deletedDate")
    name: str
    notes: str | None
    fields: list[Fields] | None = None
    reprompt: bool
    favorite: bool
    password_history: list[PasswordHistory] | None = pydantic.Field(alias="passwordHistory")


class Card(pydantic.BaseModel, extra="forbid"):
    cardholder_name: str | None = pydantic.Field(alias="cardholderName")
    brand: str | None
    number: pydantic.SecretStr | None
    exp_month: int | None = pydantic.Field(alias="expMonth")
    exp_year: int | None = pydantic.Field(alias="expYear")
    code: pydantic.SecretStr | None

    @pydantic.field_serializer("number", "code", when_used="json")
    def serialize_secretstr(self, value: pydantic.SecretStr | None) -> str | None:
        if value is None:
            return None
        return value.get_secret_value()


class ItemCard(pydantic.BaseModel, extra="forbid"):
    object: Literal["item"] = pydantic.Field(exclude=True)
    type: Literal[ItemType.card] = pydantic.Field(exclude=True)
    id: ItemID = pydantic.Field(exclude=True)
    folder_id: FolderID | None = pydantic.Field(alias="folderId")
    organization_id: OrgID | None = pydantic.Field(alias="organizationId")
    collection_ids: list[CollectionID] | None = pydantic.Field(alias="collectionIds")
    creation_date: datetime = pydantic.Field(alias="creationDate")
    revision_date: datetime = pydantic.Field(alias="revisionDate")
    deleted_date: datetime | None = pydantic.Field(alias="deletedDate")
    name: str
    card: Card = pydantic.Field(alias="card")
    notes: str | None
    fields: list[Fields] | None = None
    reprompt: bool
    favorite: bool
    password_history: list[PasswordHistory] | None = pydantic.Field(alias="passwordHistory")


class Identity(pydantic.BaseModel, extra="forbid"):
    first_name: str | None = pydantic.Field(alias="firstName")
    middle_name: str | None = pydantic.Field(alias="middleName")
    last_name: str | None = pydantic.Field(alias="lastName")
    title: str | None
    company: str | None
    email: str | None
    phone: str | None
    address1: str | None
    address2: str | None
    address3: str | None
    city: str | None
    state: str | None
    postal_code: str | None = pydantic.Field(alias="postalCode")
    country: str | None
    ssn: str | None
    username: str | None
    passport_number: str | None = pydantic.Field(alias="passportNumber")
    license_number: str | None = pydantic.Field(alias="licenseNumber")


class ItemIdentity(pydantic.BaseModel, extra="forbid"):
    object: Literal["item"] = pydantic.Field(exclude=True)
    type: Literal[ItemType.identity] = pydantic.Field(exclude=True)
    identity: Identity = pydantic.Field(alias="identity")
    id: ItemID = pydantic.Field(exclude=True)
    folder_id: FolderID | None = pydantic.Field(alias="folderId")
    organization_id: OrgID | None = pydantic.Field(alias="organizationId")
    collection_ids: list[CollectionID] | None = pydantic.Field(alias="collectionIds")
    creation_date: datetime = pydantic.Field(alias="creationDate")
    revision_date: datetime = pydantic.Field(alias="revisionDate")
    deleted_date: datetime | None = pydantic.Field(alias="deletedDate")
    name: str
    notes: str | None
    fields: list[Fields] | None = None
    reprompt: bool
    favorite: bool
    password_history: list[PasswordHistory] | None = pydantic.Field(alias="passwordHistory")


Items = Annotated[Union[ItemLogin, ItemSecureNote, ItemCard, ItemIdentity], pydantic.Field(discriminator="type")]


class ItemLoginNew(pydantic.BaseModel):
    type: Literal[ItemType.login] = ItemType.login
    name: str
    folder_id: FolderID | None = pydantic.Field(alias="folderId", default=None)
    organization_id: OrgID | None = pydantic.Field(alias="organizationId", default=None)
    collection_ids: list[CollectionID] | None = pydantic.Field(alias="collectionIds", default=None)
    login: ItemLoginData
    notes: str | None = None
    fields: list[Fields] | None = None
    reprompt: bool = False
    favorite: bool = False

    @pydantic.field_serializer("reprompt", when_used="json")
    def serialize_reprompt(self, value: bool) -> int:
        return 1 if value else 0

    @pydantic.field_validator("reprompt", mode="before")
    def validate_reprompt(cls, value: int | bool) -> bool:
        if isinstance(value, bool):
            return value
        return value == 1


# endregion



import contextlib
import logging
from typing import Any

import httpx
import pydantic

from bitwarden_rest_client.models import (
    CollectionID,
    DeleteResponse,
    Folder,
    FolderID,
    FolderNew,
    GeneratePasswordResponse,
    ItemID,
    ItemLoginNew,
    Items,
    ListResponse,
    LockResponse,
    OrgID,
    Response,
    SyncResponse,
    UnlockPayload,
    UnlockResponse,
)

_log = logging.getLogger(__name__)


class AsyncBitwardenClient:
    _client: httpx.AsyncClient

    # region Init / Dispose

    def __init__(self, base_url: str | None = None):
        if base_url is None:
            base_url = "http://localhost:8087"
        self._client = httpx.AsyncClient(base_url=base_url)

    @staticmethod
    def _payload_to_json(payload: pydantic.BaseModel | None) -> Any:
        if payload is None:
            return None
        obj = payload.model_dump(mode="json", by_alias=True, exclude_none=True)

        return obj

    @classmethod
    @contextlib.asynccontextmanager
    async def session(cls, base_url: str | None = None):
        client = cls(base_url=base_url)
        try:
            yield client
        finally:
            await client.close()

    async def close(self):
        await self._client.aclose()

    # endregion

    # region API Helpers

    async def _get[T: pydantic.BaseModel](self, cls: type[T], path: str, params: httpx.QueryParams | None = None) -> T:
        _log.debug("Params: %s", params)
        response = await self._client.get(path, params=params)
        response.raise_for_status()
        response_data = Response[cls].model_validate_json(response.text)
        if not response_data.success:
            raise RuntimeError("Request was not successful")
        return response_data.data

    async def _put[T: pydantic.BaseModel](
        self, cls: type[T], path: str, payload: pydantic.BaseModel | None = None
    ) -> T:
        response = await self._client.put(path, json=self._payload_to_json(payload))
        response.raise_for_status()
        response_data = Response[cls].model_validate_json(response.text)
        if not response_data.success:
            raise RuntimeError("Request was not successful")
        return response_data.data

    async def _post[T: pydantic.BaseModel](
        self, cls: type[T], path: str, payload: pydantic.BaseModel | None = None
    ) -> T:
        response = await self._client.post(path, json=self._payload_to_json(payload))
        response.raise_for_status()
        response_data = Response[cls].model_validate_json(response.text)
        if not response_data.success:
            raise RuntimeError("Request was not successful")
        return response_data.data

    async def _delete(self, path: str) -> bool:
        response = await self._client.delete(path)
        response.raise_for_status()
        response_data = DeleteResponse.model_validate_json(response.text)
        if not response_data.success:
            raise RuntimeError("Request was not successful")
        return response_data.success

    # endregion

    # region Lock / Unlock

    async def lock(self):
        return await self._post(LockResponse, "/lock")

    async def unlock(self, password: pydantic.SecretStr):
        payload = UnlockPayload(password=password)
        return await self._post(UnlockResponse, "/unlock", payload=payload)

    async def sync(self):
        return await self._post(SyncResponse, "/sync")

    async def generate_password(
        self,
        length: int = 20,
        uppercase: bool = True,
        lowercase: bool = True,
        numbers: bool = True,
        special: bool = False,
    ) -> pydantic.SecretStr:
        params = httpx.QueryParams()
        params = params.set("length", str(length))
        if uppercase:
            params = params.set("uppercase", "true")
        if lowercase:
            params = params.set("lowercase", "true")
        if numbers:
            params = params.set("numbers", "true")
        if special:
            params = params.set("special", "true")
        response = await self._get(GeneratePasswordResponse, "/generate", params=params)
        return response.data

    # endregion

    # region Folders

    async def folder_create(self, name: str) -> Folder:
        payload = FolderNew(name=name)
        return await self._post(Folder, "/object/folder", payload=payload)

    async def folder_update(self, folder: Folder) -> Folder:
        return await self._put(Folder, f"/object/folder/{folder.id}", payload=folder)

    async def folder_delete(self, folder: Folder) -> bool:
        return await self._delete(f"/object/folder/{folder.id}")

    async def folder_list(self, search: str | None = None) -> list[Folder]:
        params = httpx.QueryParams()
        if search is not None:
            params = params.set("search", search)
        response = await self._get(ListResponse[Folder], "/list/object/folders", params=params)
        return response.data

    async def folder_get(self, folder_id: FolderID | None) -> Folder:
        return await self._get(Folder, f"/object/folder/{folder_id}")

    # endregion

    # region Items

    async def item_create(self, item: ItemLoginNew) -> Items:
        return await self._post(Items, "/object/item", payload=item)  # type: ignore[arg-type]

    async def item_delete(self, item_id: ItemID) -> bool:
        return await self._delete(f"/object/item/{item_id}")  # type: ignore[arg-type]

    async def item_get(self, item_id: ItemID) -> Items:
        return await self._get(Items, f"/object/item/{item_id}")  # type: ignore[arg-type]

    async def item_update(self, item: Items) -> Items:
        return await self._put(Items, f"/object/item/{item.id}", payload=item)  # type: ignore[arg-type]

    async def item_list(
        self,
        org_id: OrgID | None = None,
        collection_id: CollectionID | None = None,
        folder_id: FolderID | None = None,
        url: str | None = None,
        trash: bool = False,
        search: str | None = None,
    ) -> list[Items]:
        params = httpx.QueryParams()
        if org_id is not None:
            params = params.set("organizationId", org_id)
        if collection_id is not None:
            params = params.set("collectionId", collection_id)
        if folder_id is not None:
            params = params.set("folderId", folder_id)
        if url is not None:
            params = params.set("url", url)
        if trash:
            params = params.set("trash", "true")
        if search is not None:
            params = params.set("search", search)
        response = await self._get(ListResponse[Items], "/list/object/items", params=params)
        return response.data

    # endregion



import pathlib
import tomllib

from bitwarden_rest_client import __version__


def test_version_project():
    with pathlib.Path("pyproject.toml").open("rb") as fp:
        config = tomllib.load(fp)
    assert __version__ == config["project"]["version"]


def test_version_bump():
    with pathlib.Path(".bumpversion.toml").open("rb") as fp:
        config = tomllib.load(fp)

    assert __version__ == config["tool"]["bumpversion"]["current_version"]



import enum
from typing import Annotated, cast

import pydantic
import typer
from rich.console import Console

from bitwarden_rest_client._sync.client import BitwardenClient
from bitwarden_rest_client.models import FieldHidden, ItemLoginData, ItemLoginNew, URIMatch, UriMatch

app = typer.Typer()
console = Console()


@app.command()
def lock():
    with BitwardenClient.session() as session:
        response = session.lock()
        console.print(response)


def as_secret(value: str | pydantic.SecretStr) -> pydantic.SecretStr:
    if isinstance(value, pydantic.SecretStr):
        return value
    return pydantic.SecretStr(value)


@app.command()
def unlock(
    password: Annotated[pydantic.SecretStr, typer.Option(prompt=True, hide_input=True, parser=as_secret)],
):
    with BitwardenClient.session() as session:
        response = session.unlock(password)
        console.print(response)


@app.command()
def sync():
    with BitwardenClient.session() as session:
        response = session.sync()
        console.print(response)


@app.command()
def password(
    length: Annotated[int, typer.Argument()] = 20,
    lowercase: Annotated[bool, typer.Option("--lowercase/--no-lowercase", "-l/-L")] = True,
    uppercase: Annotated[bool, typer.Option("--uppercase/--no-uppercase", "-u/-U")] = True,
    numbers: Annotated[bool, typer.Option("--numbers/--no-numbers", "-n/-N")] = True,
    special: Annotated[bool, typer.Option("--special/--no-special", "-s/-S")] = False,
):
    with BitwardenClient.session() as session:
        response = session.generate_password(
            length=length, lowercase=lowercase, uppercase=uppercase, numbers=numbers, special=special
        )
        console.print(response.get_secret_value())


# region Folder

folder_group = typer.Typer()
app.add_typer(folder_group, name="folder")


@folder_group.command("search")
def folder_search(search: Annotated[str | None, typer.Argument()] = None):
    with BitwardenClient.session() as session:
        response = session.folder_list(search=search)
        console.print(response)
        if response:
            folder = response[0]
            response = session.folder_get(folder.id)
            console.print(response)


@folder_group.command("create")
def folder_create(name: str, sync: Annotated[bool, typer.Option("--sync", "-s")] = False):
    with BitwardenClient.session() as session:
        response = session.folder_create(name=name)
        console.print(response)
        if sync:
            response = session.sync()
            console.print(response)


@folder_group.command("list")
def folder_list():
    with BitwardenClient.session() as session:
        response = session.folder_list()
        console.print(response)


# endregion

# region LoginItem

login_group = typer.Typer()
app.add_typer(login_group, name="login")


class Password(enum.StrEnum):
    PROMPT = "-"
    GENERATE = "?"


@login_group.command("create")
def login_create(
    name: str,
    username: Annotated[str | None, typer.Option("--username", "-u")] = None,
    password: Annotated[Password | None, typer.Option("--password", "-p")] = None,
    uri: str | None = None,
    folder: Annotated[str | None, typer.Option("--folder", "-f")] = None,
    sync: Annotated[bool, typer.Option("--sync", "-s")] = False,
):
    with BitwardenClient.session() as session:
        if folder is not None:
            folders = session.folder_list(search=folder)
            folders = [f for f in folders if f.name == folder]
            if not folders:
                console.print(f"[red]Folder '{folder}' not found[/red]")
                raise typer.Exit(code=1)
            if len(folders) > 1:
                console.print(f"[red]Multiple folders found with name '{folder}'[/red]")
                raise typer.Exit(code=1)
            folder_id = folders[0].id
        else:
            folder_id = None
        if uri is not None:
            uris = [
                UriMatch(match=URIMatch.exact, uri=uri),
            ]
        else:
            uris = None
        if password == Password.GENERATE:
            _password = session.generate_password()
        elif password == Password.PROMPT:
            _password = cast(
                pydantic.SecretStr, typer.prompt("Password: ", hide_input=True, value_proc=pydantic.SecretStr)
            )
        else:
            _password = None
        response = session.item_create(
            ItemLoginNew(
                name=name,
                login=ItemLoginData(
                    username=username,
                    password=_password,
                    uris=uris,
                ),
                folderId=folder_id,
            )
        )

        console.print(response)
        if sync:
            response = session.sync()
            console.print(response)


@login_group.command("update")
def login_update(name: str):
    with BitwardenClient.session() as session:
        items = session.item_list(url="ssh://root@certmgr.krypto.dk")
        item = items[0]
        if item.fields is None:
            item.fields = []
        item.fields.append(FieldHidden(name="updated", value=pydantic.SecretStr("true")))
        response = session.item_update(item)
        console.print(response)


# endregion

# region Item


@app.command()
def item(search: Annotated[str | None, typer.Argument()] = None):
    with BitwardenClient.session() as session:
        response = session.item_list(search=search)
        console.print(response)


# endregion


def main():
    app()
